
#Rem
MPPT for Ultra basic V1 Picaxe 08M2
Ver UB_08M-2
RM Software
11 Dec 2020


'todo, opto input not coded, output serial lcd data display not coded
'Mods
'removed external 12V psu, wired input volts direct to +ve out via 2 x 1N4004 diodes to drop the volts somewhat
'Run ADC at 4.096 rather than splitting between 2.048
'remove 10k resistor across RV1 to allow the higher voltage input, 
'Calibrate battery sense...
'put precision 12v across output battery terminals and set RV1 to 3.072V on cpu pin3, ie 16v max = 4.096 or 1023

Battery Volts C.4, set to 4.096v at Max output voltage = 16v
External Sync Input, C.3, normally high
PWM out C.2
Current Sense C.1, 0 to 40 Amps, 100mV\amp, 0 amps = 0.5v
Serial Out C.0, pcb link to use this


'pwmout C.2, 255, xxxxx for 32Khz 10bit pwm 0 to 1023

'ADC counts for Bulk volatge and Float
'Bulk = 14.2
'Float = 13.8
'Flat = 11

Set voltage divider pot to read 4.096 volts with 16V battery side input
this gives 1023 values = 16v
14.2 >> (14.2/16) * 1023 = 908 ADC value
13.8 >> (13.8/16) * 1023 = 882
11 >> (11/16) * 1023 = 703


FVR4096 used for current, no voltage divider in circuit
Max current = 40 amps, sensor has 100mV/Amp = 4 volts + offset, this will exceed adc 4.096
so limit to 4.096v = 1023 adc counts, 
sensor 0amps = 0.5v = 124 count, leaving 874 count for current = 35 amps
max = 874+124 = 998 counts, little bit of headroom


Battery size limiting charge current
10 amps = 1V + 0.5V offset = 1.5v = 375 count

If a charge state is interrupted, it will resume after a number of retry attempts, or otherwise revert to Night state, which restarts in a Bulk Charge
Note: charge duration timers have not been coded to resume from an interruption, they revert to startup settings.

Zero current offset will have to be entered into any calcs when displaying actual charge current on a screen
#Endrem

#picaxe 08M2
#no_data
'#no_table

'--------------------------------------------------------
'uncomment one or other
'place a led + 470R resistor between the 0v and serial out terminal on the LCD display output, led flashes on charge state
'#define UseLedStatus

'sends sertxd comments to terminal to test operation
#define UseTerminal

'external serial lcd display, no code written as I dont have a suitable display to test it.
'#define UseLCD
'--------------------------------------------------------



#define ccCurrentZeroOffset 128
#define ccNightChargeCurrent 130

'circuit limit 35 amps for mosfets and inductor
#define ccMaxAllowedChargeCurrent 1000

;determined by battery type, 15A = 500
#define ccMaxBatteryCurrent 500

'no need to test at a high current, use 2A 200
#define ccMaxQuickScanCurrent 200
		

'Bulk settings
'time interval is 0.5 seconds at M32
'1 min = 120
#define ccBulkVolts 910		
#define ccBulkChargeTimeout 120

'CV same as bulk, approx 1 hr 7200
#define ccBulkCVvolts 910
#define ccBulkCVTimeout 240

'Float settings
#define ccFloatVolts 880
#define ccFloatTimeout 500


#define ccSyncOn 0
#define ccState_Start 0
#define ccState_Bulk 1
#define ccState_BulkCV 2
#define ccState_Float 4
#define ccState_Night 8
#define ccState_Interrupted 16
#define ccState_NoCharge 32




symbol ChargeState = b0
symbol noChargeFlag = bit8
symbol errorFlag = bit9
symbol overCurrentFlag = bit10
symbol overVoltageFlag = bit11
symbol pwmPeakLocatedFlag = bit12
symbol generalTimerExpiredFlag = bit13
symbol chargeTimerStartedFlag = bit14
symbol mpptErrorFlag = bit15
symbol mpptOverVoltageFlag = bit16
symbol mpptOverCurrentFlag = bit17
symbol mpptChargeDetectedFlag = bit18
symbol mpptDisturbanceFlag = bit19
symbol mpptModeUp = bit20
symbol mpptModeDown = bit21

symbol PWMOutPin = C.2
symbol ADCCurrentSense = 1
symbol ADCBatterySense = 4
symbol ExtSyncPin = PinC.3
symbol Led = C.0

symbol pwmStart = s_w1
symbol tmpPWM = s_w2
symbol pwmMax = s_w3
symbol TimeDelayCharge = s_w4
symbol currentMax = s_w5
symbol currentRef = s_w6

symbol pwmPeak = w13
symbol LatestChargeCurrent = w12
symbol BatteryVoltage = w11
symbol tmpW_Value = w10
symbol PwmValue = w9

symbol MaxAllowedBatteryCurrent = w8
symbol MaxAllowedBatteryVoltage = w7

symbol pwmDiff = b12
symbol nightRetry_Count = b13
symbol highestCurrent = w4

symbol lookAheadCount = b6
symbol noCharge_Count = b7

symbol tmpB_Value = b4
symbol pwmStep = b5
symbol ledStatusCounter = b3

'word\byte peek,poke locations
symbol Ram_NoCharge_LastChargeState = 28
symbol Ram_SensorZeroCurrent = 30
symbol Ram_LastPWMPeak = 32



#terminal 38400


Init:
Setfreq M32
fvrsetup FVR4096
'ADC +ve ref = Internal FVR, -ve = 0v
ADCConfig %0011	
ChargeState = ccState_Start
'start pwm in background then turn off
pwmout C.2, 255, 1
pwmDuty PWMOutPin, 0
wait 8
ledStatusCounter = 0

sertxd("App Start.", cr,lf)


'run in continuous loop until the sync input opto coupler is turned on = 0
'attempt to run at MPPT max current at all times as the battery will always have dynamic loads

do Until ExtSyncPin = ccSyncOn

	Select Case ChargeState
		
Case ccState_Start
		#ifdef UseTerminal
			sertxd("Start.", cr,lf)
		#endif
		
		#ifdef UseLedStatus
			gosub BlinkStatusLed
		#endif
	
		'cal current sense for 0 current, store in ram
		pwmDuty PWMOutPin, 0
		pause 40000
		gosub ReadVoltsAmps
		poke Ram_SensorZeroCurrent, Word LatestChargeCurrent

		MaxAllowedBatteryVoltage = ccBulkVolts
		MaxAllowedBatteryCurrent = ccMaxQuickScanCurrent
		
		sertxd("Start - full scan.", cr,lf)
		gosub PVFullScan
		if noChargeFlag = 0 then
			'do any setup stuff here then set to bulk
			ChargeState = ccState_Bulk
		else
			ChargeState = ccState_NoCharge
		endif
			
		
		
Case ccState_Bulk
		#ifdef UseTerminal
			sertxd(cr,lf,"Bulk Started.", cr,lf)
		#endif
		'start the bulk phase, bring battery up to bulk voltage, hold for small timeout
		'then switch to Bulk Equalize Constant Voltage phase, some batteries this is same thing
			MaxAllowedBatteryVoltage = ccBulkVolts
			MaxAllowedBatteryCurrent = ccMaxBatteryCurrent
			TimeDelayCharge = ccBulkChargeTimeout
			gosub MpptTrack
			
			if noChargeFlag = 0 then
				ChargeState = ccState_BulkCV
			else
				ChargeState = ccState_NoCharge
			endif
			
			
			
	Case ccState_BulkCV
		'start the bulk constant voltage equalize phase
		'hold for 1 - 6 hours, depends on battery
			#ifdef UseTerminal
				sertxd(cr,lf,"Bulk CV.",cr,lf)
			#endif
			MaxAllowedBatteryCurrent = ccMaxBatteryCurrent
			MaxAllowedBatteryVoltage = ccBulkCVVolts
			TimeDelayCharge = ccBulkCVTimeout
			gosub MpptTrack
			
			if noChargeFlag = 0 then
				'cv phase done, move to Float
				ChargeState = ccState_Float
			else
				ChargeState = ccState_NoCharge
			endif
			
			
			
			
	Case ccState_Float
		'start float, voltage point lowered, note still in mppt mode so charge maintained until PV runs out
		'timeout is max ie 9 hours, ie be into night 
			#ifdef UseTerminal
				sertxd(cr,lf,"Float.",cr,lf)
			#endif
		
			MaxAllowedBatteryCurrent = ccMaxBatteryCurrent
			MaxAllowedBatteryVoltage = ccFloatVolts
			TimeDelayCharge = ccFloatTimeout
			'the bulk voltage is higher, so there will be no charge until battery voltage drops
			'wait before calling charge again with float settings
			#ifdef UseTerminal
				sertxd(cr,lf,"Float waiting..",cr,lf)
			#endif
			gosub WaitForBatteryVoltsToDrop
			
			gosub MpptTrack
			
			if noChargeFlag = 0 then
				'float timeout done, but we always stay in float so repeat
				ChargeState = ccState_Float
			else
				ChargeState = ccState_NoCharge
			endif
		
			
			
	Case ccState_NoCharge
		'no charge state occurred, ie PV fuse blown, dark cloud, no sun etc
			#ifdef UseTerminal
				sertxd("No Charge State.", cr,lf)
			#endif
		
			MaxAllowedBatteryVoltage = ccFloatVolts
			MaxAllowedBatteryCurrent = ccMaxQuickScanCurrent
			
			gosub PVFullScan
			if noChargeFlag = 0 then
			'PV present
				ChargeState = ccState_Start
			else
				'still nothing, set to night
				ChargeState = ccState_Night
			endif
			
		
			
	Case ccState_Night
		'wait for a period and retry for charge current
		'if present then set to start
		#ifdef UseTerminal
			sertxd("Night.", cr,lf)
		#endif
	
		MaxAllowedBatteryVoltage = ccFloatVolts
		MaxAllowedBatteryCurrent = ccMaxQuickScanCurrent
		
		inc nightRetry_Count
		pause 8000
		if nightRetry_Count = 0 then
			
			#ifdef UseTerminal
				sertxd("Night Check PV", cr,lf)
			#endif
			gosub PVFullScan
			if noChargeFlag = 0 then
			'PV present
				ChargeState = ccState_Start
			endif
		
		endif
	
	
	
		#ifdef UseLedStatus
			gosub BlinkStatusLed
		#endif
	
	
	Case ccState_Interrupted
		'single opto isolated input, if activated by interrupt on pin C.3 then do something here
		#ifdef UseTerminal
			sertxd("Opto Input", cr,lf)
		#endif
		
	EndSelect


Loop


SyncOnActivated:
'turn off pwm
pwmduty PWMOutPin, 0

End

'---------------------------------------------------------------------------------------------------

MpptTrack:
'pass in pwmPeak from PVFullScan, routine works for all main charge states
'Bulk, Constant Voltage CV and Float
#ifdef UseTerminal
	sertxd("Mppt Track Start PWM:", #pwmPeak, cr,lf)
#endif

'clear flag, will be set on overvoltage
mpptOverVoltageFlag = 0
'set after charge time delay expired
generalTimerExpiredFlag = 0
'once voltage trip point reached then start a charge duration timer
chargeTimerStartedFlag = 0

'set initial level lower to allow soft start
pwmValue = pwmPeak min 100 - 50
lookAheadCount = 1
pwmDiff = 25

do
'set new pwm value, look at error flags and loop back to reduce voltage or current in error state
'no errors, depending on the charge current look ahead to locate a mppt PV point, low charge currents < 1 amp
'require a much higher look ahead value.

	gosub SetPWM
	
	if errorFlag = 0 then
	'do the mppt track if no errors
			
			if lookAheadCount = 1 then
				'take pwm and current snapshot
				currentRef = LatestChargeCurrent
				tmpPWM = pwmValue
				#ifdef UseTerminal
					sertxd("Ref Snapshot PWM:", #pwmValue, "  Amps:", #LatestChargeCurrent,cr,lf)
				#endif
				
			elseif lookAheadCount = pwmDiff then
				'compare to snapshot, if current less then restore, else continue increasing pwm
				if LatestChargeCurrent <= currentRef then
					'- restore, subtract last difference -2
					pwmValue = tmpPWM min 130 - pwmDiff - 3
					#ifdef UseTerminal
						sertxd("Diff PWM:", #pwmValue, "  Amps:", #LatestChargeCurrent,cr,lf)
					#endif
				endif
				
				'allow new snapshot
				lookAheadCount = 0
				
				'bigger the current more sensitive to difference, so reduce the diff value
				'values here work well for charge currents in the milliamps to 20A + reqion
				'on the 300W PV 32V panel and 12V battery used for testing
				select Case currentRef
				Case <= 150
					pwmDiff = 25
				Case 151 to 250
					pwmDiff = 6
				Case 251 to 350
					pwmDiff = 3
				Case > 350
					pwmDiff = 2
				end Select
				
			endif
		
			inc lookAheadCount
			pwmValue = pwmValue max 850 + 1 
			'100% pwm = 1023, lower voltage panels will require a higher max limit
			
						
	endif

	


'look at various global flags and starts time countdown for charge duration timers
'these are set when a voltage end point is reached, eg once bulk reaches its set voltage then start a timer for CV etc
'if the high battery voltage flag set then setpoint has been reached for charge duration
	if mpptOverVoltageFlag = 1 AND chargeTimerStartedFlag = 0 then
		'start charge timer
		chargeTimerStartedFlag = 1
		Time = 0
		#ifdef UseTerminal
			sertxd("Charge Timer Started: ", #TimeDelayCharge, cr,lf)
		#endif
		
	endif
	
	if chargeTimerStartedFlag = 1 AND Time > TimeDelayCharge then
	'timeout expired
		generalTimerExpiredFlag = 1
		#ifdef UseTerminal
			sertxd("Charge Timer Ended: ", cr,lf)
		#endif
	endif


'if sun goes down or PV breaker pops then need a way to exit the loop on No Charge timeout
'if delay expired then set flag and exit, 250 allows a big enough pwm value to get some charge current
	if currentRef < ccNightChargeCurrent then
		inc noCharge_Count
		if noCharge_Count > 250 then
			noChargeFlag = 1
		endif
	else
		noCharge_Count = 0 
	endif



loop until noChargeFlag = 1 OR generalTimerExpiredFlag = 1

'turn off pwm, this turns off the PV isolation Mosfet
pwmDuty PWMOutPin, 0
Pause 60000

Return
'------------------------------------------------------------------------------------------------------------
	
	
'-------------------------------------------------------------------------------------------------------------	
SetPWM: '(pwmValue)
'Uses pwmValue, set the duty cycle and read the Battery voltage and Charge current
	
	'this delay is required as the current analog filter has a slow response, so allow time for stability
	'after a change in pwm, will update the filter components and return to this.
	'The ADC's dont require multiple reads as they have active LP filters.
	
	pwmDuty PWMOutPin, pwmValue
	pause 500
	
	readADC10 ADCCurrentSense, LatestChargeCurrent
	readADC10 ADCBatterySense, BatteryVoltage
		
		
	'test for over voltage or current, error flags reset - then set on error
	gosub TestForOverVoltageCurrent
		
	if overVoltageFlag = 1 then
	'reduce pwm a lot
		pwmValue = pwmValue MIN 150 - 30
		#ifdef UseTerminal
			sertxd("Over Volts - PWM:",#pwmValue, cr,lf )
		#endif
		
		mpptOverVoltageFlag = 1
		'set mppt over voltage flag, this will be used to start any charge timers
		
	endif
	
	
	if overCurrentFlag = 1 then
		'reduce pwm a little less
		pwmValue = pwmValue MIN 150 - 10
		#ifdef UseTerminal
			sertxd("Over Current - PWM:",#pwmValue, cr,lf )
		#endif
	endif
	
	
	
Return	
	
	
	
	
	
	
	
	
	
	
	

'-------------------------------------------------------------------------------------------
WaitForBatteryVoltsToDrop:
'After a bulk charge and moving to float, call this routine prior to running float
'wait for the battery voltage to drop, otherwise a zero charge could result
'In the real world with a loaded battery, this doesnt happen.
	do 
		gosub ReadVoltsAmps
		#ifdef UseTerminal
			sertxd("Waiting Voltage:",#BatteryVoltage, "  Max:", #MaxAllowedBatteryVoltage, cr,lf)	
		#endif
		pause 1000
	Loop Until BatteryVoltage < MaxAllowedBatteryVoltage
Return
'-------------------------------------------------------------------------------------------






'-------------------------------------------------------------------------------------------
PVFullScan:

errorFlag = 0
pwmPeakLocatedFlag = 0
noChargeFlag = 0
highestCurrent = 0

#ifdef UseTerminal
	sertxd("Full Scan Start:",cr,lf)
#endif


'pwm range typically 250 to 750, depends on PV voltage and battery
For PwmValue = 50 to 900 step 10
	pwmDuty PWMOutPin, pwmValue
	pause 500
	gosub ReadVoltsAmps	
	'LatestChargeCurrent, BatteryVoltage are Updated
	
	gosub TestForOverVoltageCurrent
	'flag set on over current or voltage
	if errorFlag = 1 then
		'reduce the pwm to a lower safe level
		tmpW_Value = pwmValue min 50 - 30
		pwmDuty PWMOutPin,tmpW_Value
		pause 500
	endif
	
	
	'locate a PV peak current
	if LatestChargeCurrent > highestCurrent then
		highestCurrent = LatestChargeCurrent
		pwmPeak = PWMValue
	endif

'get difference in current measurements
'see if difference is getting bigger, ie current reducing
	tmpW_Value = highestCurrent - LatestChargeCurrent
	if tmpW_Value > 2 OR errorFlag = 1 then
		'past peak of panel or over current\voltage
		pwmPeakLocatedFlag = 1
		pwmMax = PWMValue
		exit
	endif


next

'if the pwmPeakLocatedFlag is zero no pv peak found
'if no charge then something wrong
if highestCurrent < ccNightChargeCurrent then
	noChargeFlag = 1
	pwmPeak = 50
endif

#ifdef UseTerminal
	sertxd("Full Scan Done: Pk: ", #pwmPeak, "  Max:", #pwmMax, "  Pk Located:", #pwmPeakLocatedFlag,"  No Charge Flag:",#noChargeFlag, cr,lf)
#endif

'turn off pwm, this turns off the PV isolation Mosfet
pwmDuty PWMOutPin, 0
pause 60000

return

'--------------------------------------------------------------------------------------------



#ifdef UseLedStatus
BlinkStatusLed:

inc ledStatusCounter
select case ChargeState
		
	case ccState_Start
	high Led
	pause 16000
	low Led
		
case ccState_Bulk
	'max current mode upto a set voltage
	tmpA_Value = ledStatusCounter //4
	if tmpA_Value  = 0 then
		pulsout Led, 800
	endif


case ccState_Bulk, ccState_BulkCV
	'reached the set voltage, held at that value with decreasing current for a set time
	tmpA_Value = ledStatusCounter //10
	if tmpA_Value  = 0 then
		pulsout Led, 800
	endif
	
		
case ccState_Float
	'lower float voltage
	tmpA_Value = ledStatusCounter //30
	if tmpA_Value  = 0 then
		pulsout Led, 1000
	endif
		
case ccState_night
	'no charge waiting
	pulsout Led, 8000
	
		
endSelect



Return
#endif



'-------------------------------------------------------------------------------------------
TestForOverVoltageCurrent:
'set error flag on over voltage or current
'set overVoltageFlag or overCurrentFlag also

errorFlag = 0
overVoltageFlag = 0
overCurrentFlag = 0

	if BatteryVoltage > MaxAllowedBatteryVoltage then
		#ifdef UseTerminal then
			sertxd("Range OV: ", #BatteryVoltage, cr,lf)
		#endif
		overVoltageFlag = 1
		
		
	elseif LatestChargeCurrent > ccMaxAllowedChargeCurrent OR  LatestChargeCurrent > MaxAllowedBatteryCurrent then
	'Hardware or Battery specific limit exceeded
		overCurrentFlag = 1
		
		#ifdef UseTerminal then
			sertxd("Range OC: ", #LatestChargeCurrent,  cr,lf)
		#endif
	endif

	errorFlag = overVoltageFlag OR overCurrentFlag
	return
'-------------------------------------------------------------------------------------------



'--------------------------------------------------------------------------------------------------
ReadVoltsAmps:
'read current sensor and Battery voltage 
'
	readADC10 ADCCurrentSense, LatestChargeCurrent
	readADC10 ADCBatterySense, BatteryVoltage

Return

'--------------------------------------------------------------------------------------------------	




	
	
	
	
	
	
	
	
	
	
	
	
	